from sssekai.unity.AssetBundle import load_assetbundle
import os, json, tqdm


def main_mvdata(args):
    from UnityPy.enums import ClassIDType

    source = args.indir
    source = os.path.expanduser(source)
    source = os.path.abspath(source)
    outfile = os.path.abspath(args.outfile)
    os.chdir(source)
    mvdata_keys = sorted(os.listdir(source))
    mvdata_items = list()
    for key in tqdm.tqdm(mvdata_keys):
        try:
            with open(key, "rb") as f:
                env = load_assetbundle(f)
                for obj in env.objects:
                    if obj.type == ClassIDType.MonoBehaviour:
                        data = obj.read()
                        typetree = obj.read_typetree()
                        typetree = {
                            k: v for k, v in typetree.items() if not k.startswith("m_")
                        }
                        if "name" in typetree:
                            mvdata_items.append(typetree)
        except Exception as e:
            print(f"skipping {key}: {e}")
    outdir = os.path.dirname(outfile)
    if outdir:
        os.makedirs(outdir, exist_ok=True)
    with open(outfile, "w", encoding="utf-8") as f:
        json.dump(mvdata_items, f, indent=4, ensure_ascii=False)



import os
from pathlib import Path
from logging import getLogger

logger = getLogger('abdecrypt')

def main_abdecrypt(args):
    from sssekai.crypto.AssetBundle import decrypt_iter

    args.outdir = Path(os.path.abspath(args.outdir))
    args.indir = Path(os.path.abspath(args.indir))
    args.outdir.mkdir(parents=True, exist_ok=True)
    tree = os.walk(args.indir)
    assert args.indir != args.outdir, "Input and output directories must be different"
    for root, dirs, files in tree:
        for fname in files:
            file = Path(root) / fname
            if file.is_file():                
                with open(file, "rb") as src:
                    next_bytes = lambda nbytes: src.read(nbytes)
                    out_path = args.outdir / file.relative_to(args.indir)
                    out_path.parent.mkdir(parents=True, exist_ok=True)
                    logger.info("Decrypting %s -> %s", file.as_posix(), out_path.as_posix())
                    with open(out_path, "wb") as dest:
                        for block in decrypt_iter(next_bytes):
                            dest.write(block)



import logging, os, shutil

logger = logging.getLogger(__name__)


METADATA_MAGIC = b"\xaf\x1b\xb1\xfa"
SIG_DW_OP_piece = b"DW_OP_piece not implemented\x00x5\x00x10\x00"


def main_il2cpp(args):
    try:
        import lief
    except ImportError as e:
        logger.error(
            "Please install sssekai[il2cpp] through your Python package manager to use the il2cpp entrypoint"
        )
        raise e
    logger.info("Loading metadata from %s", args.metadata)
    metadata = bytearray(open(args.metadata, "rb").read())
    assert (
        metadata[:4] != METADATA_MAGIC
    ), "Metadata file is unlikely to be protected with XOR cipher (non-obfuscated header)"
    logger.info("Loading binary from %s", args.binary)
    elf = lief.ELF.parse(args.binary)
    text: lief.ELF.Section = elf.get_section(".rodata")
    text = text.content.tobytes()
    # Yes - the heuristic is *that* simple.
    # We're probably fine if Unity never changes their bundled NDK for JP
    # Verified to work with JP 3.8.1 & 5.5.0
    off = text.find(SIG_DW_OP_piece)
    assert (
        off != -1
    ), "Signature not found in the binary. NOTE: Only JP Android releases are verified and supported."
    off += len(SIG_DW_OP_piece)
    key = text[off : off + 0x80]
    logger.info("Key: %s", key.hex())
    for i in range(0, len(metadata)):
        metadata[i] ^= key[i & 0x7F]
        if i == 4:
            assert metadata[:4] == METADATA_MAGIC, "Metadata deobfuscation failed"
    logger.info("Metadata seems OK. Saving files to %s", args.outdir)
    os.makedirs(args.outdir, exist_ok=True)
    with open(os.path.join(args.outdir, "global-metadata.dat"), "wb") as f:
        f.write(metadata)
    shutil.copy(args.binary, os.path.join(args.outdir, "il2cpp.so"))



from . import *


def test():
    import UnityPy
    from UnityPy.enums import ClassIDType
    from sssekai.unity.AnimationClip import AnimationHelper

    file = sample_file_path("animation", "legacy")
    env = UnityPy.load(file)
    anim = next(
        filter(lambda x: x.type == ClassIDType.AnimationClip, env.objects)
    ).read()
    helper = AnimationHelper.from_clip(anim)
    pass


if __name__ == "__main__":
    test()



import logging

logger = logging.getLogger(__name__)

DEFAULT_SEKAI_UNITY_VERSION = "2022.3.21f1"
_sssekai_unity_version = DEFAULT_SEKAI_UNITY_VERSION


def sssekai_get_unity_version():
    return _sssekai_unity_version


def sssekai_set_unity_version(value):
    global _sssekai_unity_version
    if _sssekai_unity_version != value:
        logger.warning(
            f"Setting Unity Version from {_sssekai_unity_version} to {value}"
        )
        _sssekai_unity_version = value



from . import AbCache, SekaiUserData, fromdict
from logging import getLogger

logger = getLogger(__name__)


def register_as_anonymous_user_sega(cache: AbCache) -> SekaiUserData:
    """Registers an anonymous account on SEGA servers, and writes relavant data to the AbCacheobject.

    Args:
        cache: The AbCache object to write the user data to.

    Returns:
        SekaiUserData: The user data object.
    """
    logger.info("Registering user data")
    # https://github.com/mos9527/sssekai/issues/31
    cache._update_request_headers()
    cache.update_signatures()
    payload = {
        "platform": cache.headers["X-Platform"],
        "deviceModel": cache.headers["X-DeviceModel"],
        "operatingSystem": cache.headers["X-OperatingSystem"],
    }
    resp = cache.request_packed("POST", cache.SEKAI_API_USER, data=payload)
    data = cache.response_to_dict(resp)
    cache.config.auth_credential = data["credential"]
    logger.info("Success. User ID=%s" % cache.config.auth_userID)
    cache.database.sekai_user_data = data
    return fromdict(SekaiUserData, data)



from sssekai.unity.AssetBundle import load_assetbundle
from UnityPy.enums import ClassIDType
from os import path, remove, makedirs
from logging import getLogger

logger = getLogger(__name__)


def main_usmdemux(args):
    with open(args.infile, "rb") as f:
        env = load_assetbundle(f)
        datas = dict()
        for obj in env.objects:
            if obj.type in {ClassIDType.MonoBehaviour, ClassIDType.TextAsset}:
                data = obj.read()
                datas[data.m_Name] = data
        movieInfo = datas.get("MovieBundleBuildData", None)
        assert movieInfo, "Invalid AssetBundle. No MovieBundleBuildData found!"
        # movieInfo = movieInfo.read_typetree()
        usm_name = movieInfo.movieBundleDatas[0].usmFileName[: -len(".bytes")]
        logger.info("USM: %s" % usm_name)        
        usm_out = path.abspath(args.outfile)
        makedirs(path.dirname(usm_out), exist_ok=True)        
        with open(usm_out, "wb") as usmstream:
            for data in movieInfo.movieBundleDatas:
                usm = data.usmFileName[: -len(".bytes")]
                usm = datas[usm]
                usmstream.write(usm.m_Script.encode("utf-8", "surrogateescape"))
        logger.info("Saved raw USM file to %s" % usm_out)



# fmt: off
# Auto-generated by https://github.com/mos9527/UnityPyTypetreeCodegen
from typing import List, Union, Optional, TypeVar, Type
from UnityPy.files.ObjectReader import ObjectReader
from UnityPy.classes import *
from UnityPy.classes.math import (ColorRGBA, Matrix3x4f, Matrix4x4f, Quaternionf, Vector2f, Vector3f, Vector4f, float3, float4,)

UTTCG_Classes = dict()
def UTTCGen(fullname: str, typetree: dict):
    """dataclass-like decorator for typetree classess with nested type support
    
    limitations:
    - the behavior is similar to slotted dataclasses where shared attributes are inherited
      but allows ommiting init of the parent if kwargs are not sufficient
    - generally supports nested types, however untested and could be slow	
    - and ofc, zero type checking and safeguards :/	
    """    
    REFERENCED_ARGS = {'object_reader'}
    def __inner(clazz: T) -> T:
        # Allow these to be propogated to the props
        def __init__(self, **d):        
            def reduce_init(clazz, **d):
                types : dict = clazz.__annotations__
                for k, sub in types.items():
                    if type(sub) == str:
                        sub = eval(sub) # attrs turns these into strings...why?
                    while sub.__name__ == "Optional":
                        sub = sub.__args__[0]  # Reduce Optional[T] -> T
                    reduce_arg = getattr(sub, "__args__", [None])[0]
                    if k in REFERENCED_ARGS: # Directly refcounted
                        reduce_arg = sub = lambda x: x                         
                    if isinstance(d[k], list):
                        if hasattr(reduce_arg, "__annotations__"):
                            setattr(self, k, [reduce_arg(**x) for x in d[k]])
                        else:
                            setattr(self, k, [reduce_arg(x) for x in d[k]])
                    elif isinstance(d[k], dict) and hasattr(sub, "__annotations__"):
                        setattr(self, k, sub(**d[k]))
                    else:
                        if isinstance(d[k], dict):
                            setattr(self, k, sub(**d[k]))
                        else:
                            setattr(self, k, sub(d[k]))
            def reduce_base(clazz, **d):	
                for __base__ in clazz.__bases__:
                    if hasattr(__base__, "__annotations__"):
                        types : dict = __base__.__annotations__
                        args = {k:d[k] for k in types if k in d}
                        if len(args) == len(types):
                            super(clazz, self).__init__(**args)
                            reduce_init(__base__, **d)                       
                    reduce_base(__base__, **d)
            reduce_base(clazz, **d)               
            reduce_init(clazz, **d)            
        def __repr__(self) -> str:
            return f"{clazz.__name__}({', '.join([f'{k}={getattr(self, k)!r}' for k in self.__annotations__])})"
        def __save(self):
            self.object_reader.save_typetree(self, self.__typetree__)
        clazz.__init__ = __init__
        clazz.__repr__ = __repr__
        clazz.__typetree__ = typetree
        clazz.__fullname__ = fullname
        clazz.save = __save
        UTTCG_Classes[fullname] = clazz
        return clazz
    return __inner

# Helper functions
def UTTCGen_GetClasss(src: MonoBehaviour | str) -> Type:
    """Get the class definition from MonoBehaviour or a full type name."""
    if isinstance(src, MonoBehaviour):
        script = src.m_Script.read()
        src = script.m_ClassName
        if script.m_Namespace:
            src = f"{script.m_Namespace}.{src}"    
    return UTTCG_Classes.get(src, None)

T = TypeVar("T")
def UTTCGen_AsInstance(cls : Type[T], src: MonoBehaviour | ObjectReader) -> T:
    """Instantiate a class from the typetree definition and the raw data.

    In most cases, this is the function you want to use.
    It will read the typetree data from the MonoBehaviour instance and instantiate the class with the data.

    Args:
        cls: The class to instantiate. This should be a class that has been decorated with the UTTCGen decorator.
        src (MonoBehaviour | ObjectReader): The MonoBehaviour instance or ObjectReader to read from.        

    Returns:
        An instance of the class defined by the typetree.
    """
    if isinstance(src, MonoBehaviour):
        src = src.object_reader
    raw_def = src.read_typetree(cls.__typetree__, check_read=False)
    instance = cls(object_reader=src, **raw_def)
    return instance



import zipfile
import UnityPy
import logging
import json
import sys

from io import BytesIO

from UnityPy.enums import ClassIDType
from UnityPy.helpers.Tpk import get_typetree_node
from sssekai.unity.AssetBundle import load_assetbundle
from tqdm import tqdm

REGION_MAP = {
    # NOTE: Order is used to determine the region
    "com.hermes.mk": "cn",
    "com.hermes.mk.asia": "tw",
    "com.sega.ColorfulStage.en": "en",
    "com.sega.pjsekai": "jp",
    "com.pjsekai.kr": "kr",
}
ROW_REGIONS = {"cn", "tw", "kr"}
logger = logging.getLogger("apphash")

PROD_BUNDLES = {
    "6350e2ec327334c8a9b7f494f344a761",  # PJSK Android
    "c726e51b6fe37463685916a1687158dd",  # PJSK iOS
    "data.unity3d",  # TW,KR,CN (ByteDance)
}


def enum_candidates(zip_file, filter):
    return (
        (f, zip_file.open(f), zip_file) for f in zip_file.filelist if filter(f.filename)
    )


def enum_package(zip_file):
    yield zip_file
    for f in zip_file.filelist:
        if f.filename.lower().endswith(".apk"):
            yield zipfile.ZipFile(zip_file.open(f))


def dump_axml_stringpool(f: BytesIO):
    read_int = lambda nbytes: int.from_bytes(f.read(nbytes), "little")
    f.seek(8)
    hdr_type, hdr_size, size = read_int(2), read_int(2), read_int(4)
    n_strings = read_int(4)
    unk = read_int(4)
    flags = read_int(4)
    is_utf8 = flags & (1 << 8)
    string_offset = read_int(4) + 8
    unk = read_int(4)
    offsets = [read_int(4) for _ in range(n_strings)]
    for offset in offsets:
        f.seek(string_offset + offset)
        string_len = read_int(2)
        if is_utf8:
            string = f.read(string_len).decode("utf-8")
        else:
            string = f.read(string_len * 2).decode("utf-16-le")
        yield string


def dump_metadata_stringpool(f: BytesIO):
    metadata = f.read()
    metadata = metadata[metadata.find(b"\xaf\x1b\xb1\xfa") :]
    f = BytesIO(metadata)
    read_int = lambda nbytes: int.from_bytes(f.read(nbytes), "little")
    assert read_int(4) == 0xFAB11BAF, "bad sanity"
    logger.info("metadata version: %s" % read_int(4))
    stringLiteralOffset = read_int(4)  # string data for managed code
    stringLiteralSize = read_int(4)
    stringLiteralDataOffset = read_int(4)
    stringLiteralDataSize = read_int(4)
    f.seek(stringLiteralOffset)
    stringLiterals = [
        (read_int(4), read_int(4)) for _ in range(stringLiteralSize // 8)
    ]  # length, dataIndex
    for length, dataIndex in stringLiterals:
        f.seek(stringLiteralDataOffset + dataIndex)
        string = f.read(length)
        yield string.decode("utf-8")
    pass


def main_apphash(args):
    UnityPy.config.SERIALIZED_FILE_PARSE_TYPETREE = False
    env = UnityPy.Environment()
    app_package = None
    app_version = "unknown"
    app_hash = "unknown"
    app_metadata_strings = None
    proxies = None
    if args.proxy:
        logger.info("Overriding proxy: %s", args.proxy)
        proxies = {"http": args.proxy, "https": args.proxy}
    if not args.ab_src:
        if not args.apk_src or args.fetch:
            from requests import get

            src = BytesIO()
            logger.debug("Fetching latest game package (JP) from APKPure")
            resp = get(
                "https://d.apkpure.net/b/XAPK/com.sega.pjsekai?version=latest",
                stream=True,
                proxies=proxies,
            )
            size = resp.headers.get("Content-Length", -1)
            with tqdm(total=int(size), unit="B", unit_scale=True) as progress:
                for chunk in resp.iter_content(chunk_size=2**20):
                    src.write(chunk)
                    progress.update(len(chunk))
            src.seek(0)
        else:
            src = open(args.apk_src, "rb")
        with zipfile.ZipFile(src, "r") as zip_ref:
            manifests = [
                manifest
                for package in enum_package(zip_ref)
                for manifest in enum_candidates(
                    package, lambda fn: fn == "AndroidManifest.xml"
                )
            ]
            manifest = manifests[0][1]
            manifest_strings = set(dump_axml_stringpool(manifest))
            # Heur: Reverse lookup the package name from the manifest strings
            for ky in REGION_MAP:
                if ky in manifest_strings:
                    app_package = ky
                    break
            # if REGION_MAP[app_package] in ROW_REGIONS:
            #     metadata = [
            #         meta
            #         for package in enum_package(zip_ref)
            #         for meta in enum_candidates(
            #             package,
            #             lambda fn: fn.endswith("global-metadata.dat"),
            #         )
            #     ]
            #     if metadata:
            #         metadata = metadata[0][1]
            #         metadata_strings = list(dump_metadata_stringpool(metadata))
            #     app_metadata_strings = [s for s in metadata_strings if "bytedgame" in s]
            #     pass
            candidate_filter = lambda fn: fn.split("/")[-1] in PROD_BUNDLES
            candidate_filter_deep = lambda fn: ("_/" + fn).split("/")[-2] == "Data"

            candidates = [
                candidate
                for package in enum_package(zip_ref)
                for candidate in enum_candidates(
                    package,
                    candidate_filter_deep if args.deep else candidate_filter,
                )
            ]
            for candidate, stream, _ in tqdm(candidates, desc="Loading"):
                env.load_file(stream.read())
    else:
        logger.info("Loading from AssetBundle %s" % args.ab_src)
        with open(args.ab_src, "rb") as f:
            env = load_assetbundle(BytesIO(f.read()))

    from sssekai.generated import UTTCGen_AsInstance
    from sssekai.generated.Sekai import (
        AndroidPlayerSettingConfig,
        IOSPlayerSettingConfig,
    )

    res = dict()
    for reader in tqdm(env.objects, desc="Processing"):
        if reader.container:
            logger.info("Processing %s (%s)", reader.container, reader.type)
        if reader.type == UnityPy.enums.ClassIDType.MonoBehaviour:
            mono = reader.read(check_read=False)
            clazz = None
            platform = None
            if "_android" in mono.m_Name:
                clazz = AndroidPlayerSettingConfig
                platform = "android"
            elif "_ios" in mono.m_Name:
                clazz = IOSPlayerSettingConfig
                platform = "ios"
            if clazz:
                # Works with post 3.4 (JP 3rd Anniversary) builds and downstream regional builds
                try:
                    config = UTTCGen_AsInstance(clazz, reader)
                except Exception as e:
                    logger.error("Failed to parse config for %s: %s", mono.m_Name, e)
                    continue
                config: AndroidPlayerSettingConfig | IOSPlayerSettingConfig
                app_version = "%s.%s.%s" % (
                    config.clientMajorVersion,
                    config.clientMinorVersion,
                    config.clientBuildVersion,
                )
                data_version = "%s.%s.%s" % (
                    config.clientDataMajorVersion,
                    config.clientDataMinorVersion,
                    config.clientDataBuildVersion,
                )
                ab_version = "%s.%s.%s" % (
                    config.clientMajorVersion,
                    config.clientMinorVersion,
                    config.clientDataRevision,
                )
                app_hash = config.clientAppHash
                # XXX: CN build has the same package name as the JP build??
                # We need another heuristic to determine the package name
                # This is previously done by setting app_package
                package_heur = app_package or config.bundleIdentifier
                region = region = REGION_MAP.get(package_heur, "unknown")
                print(
                    f"Found {config.productName} at {config.m_Name}",
                    f"  Memo: {config.memo}",
                    f"  Package: {config.bundleIdentifier} (actually assumed as {package_heur})",
                    f"  Platform: {platform}",
                    f"  AppHash (app_hash):     {config.clientAppHash}",
                    f"  Region  (app_region):   {region} (determined by {package_heur})",
                    f"  Version (app_version):  {app_version}",
                    f"  Version (ab_version):   {ab_version}",
                    f"  Bundle Version: {config.bundleVersion}",
                    f"  Data Version:   {data_version}",
                    f"  Version Suffix: {config.clientVersionSuffix}",
                    "",
                    "",
                    sep="\n",
                    file=sys.stderr,
                )
                app_package = app_package or "Unknown Package (Failed APK Heuristic)"
                # fmt: off
                match args.format:
                    case 'json':
                        res[mono.m_Name] = {
                            "package": app_package,
                            "reported_package": config.bundleIdentifier,
                            "app_hash": app_hash,
                            "app_region": region,
                            "app_version": app_version,
                            "app_platform": platform,
                            "ab_version": ab_version,
                        }
                    case "markdown":
                        res[mono.m_Name] = f"""{app_package} ({app_version}, {region})
---
Reported Package: {config.bundleIdentifier}

|{'app_hash'.rjust(48)}|   app_region|  app_version|   ab_version|
|{'-'.rjust(48, '-')}|-------------|-------------|-------------|
|{app_hash.rjust(48)}|{region.rjust(13)}|{app_version.rjust(13)}|{ab_version.rjust(13)}|

- CLI Usage:

        sssekai abcache --app-platform {platform} --app-region {region} --app-version {app_version} --app-appHash {app_hash} --app-abVersion {ab_version}

- Python Usage:

        from sssekai.abcache import AbCacheConfig

        AbCacheConfig(
            app_region="{region}",
            app_version="{app_version}",
            ab_version="{ab_version}",
            app_hash="{app_hash}",
            app_platform="{platform}"
        )
"""
    print("###### RESULTS ######", file=sys.stderr)
    res = dict(sorted(res.items(), key=lambda x: x[0]))
    match args.format:
        case "json":
            print(json.dumps(res, indent=4, ensure_ascii=False))
        case "markdown":
            for name, content in res.items():
                print(f"## {name}\n{content}\n")



# Test with:
# pytest -v -s .\tests\_test_abcache.py
from . import *
import traceback

CONFIG = "https://raw.githubusercontent.com/mos9527/sekai-apphash/refs/heads/master/%(region)s/apphash.json"
REGIONS = ["en", "cn", "jp", "tw", "kr"]

from requests import get


def test_abcache():
    from sssekai.entrypoint.abcache import main_abcache

    for region in REGIONS:
        flag = False
        for _ in range(5):
            try:
                config = get(CONFIG % {"region": region}).json()
                config = config.get("production_android", config)
                main_abcache(
                    NamedDict(
                        {
                            "app_platform": "android",
                            "app_region": region,
                            "app_version": config["app_version"],
                            "app_appHash": config["app_hash"],
                            "app_abVersion": config["ab_version"],
                            "db": os.path.join(TEMP_DIR, "abcache.db"),
                        }
                    )
                )
                flag = True
                break
            except Exception as e:
                logger.error(f"Error occurred for region {region}: {e}")
                logger.error(traceback.format_exc())
                logger.error("Retrying. n=%d" % _)
        assert flag, f"All attemptes failed for region {region}"


if __name__ == "__main__":
    test_abcache()



import os, logging, datetime, time, sys, math
from shutil import copyfileobj
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer
from sssekai import __version__
from sssekai.abcache.fs import AbCacheFilesystem

logger = logging.getLogger("abserve")
fs: AbCacheFilesystem = None

def filesize(size):
    suffixes = ["B", "KB", "MB", "GB", "TB"]
    exp = int(math.floor(math.log(size, 1 << 10)))
    return f"{size / (1 << (10 * exp)):.2f} {suffixes[exp]}"
class AbServeHTTPRequestHandler(BaseHTTPRequestHandler):
    ENCODING = "utf-8"

    def format_listing(self, path):
        t0 = time.time()
        r = []
        path = path or "/"
        title = f"Directory listing for {path}"
        r = (
            f"<!DOCTYPE HTML>"
            f'<html lang="en">'
            f"<head>"
            f'<meta charset="utf-8">'
            f"<style>"
            f"body {{ font-family: monospace; }}"
            f"body {{ background-color: black; color: white; }}"
            f"a,i {{ color: lightblue; }}"
            f"</style>"
            f"<title>{title}</title>"
            f"</head>"
            f"<body><h1>{title}</h1>"
            f"<i>children: {len(fs.ls(path))},</i>"
            f"<i>total number of files: {fs.info(path)['file_count']},</i>"
            f"<i>total size: {filesize(fs.info(path)['total_size'])}</i><br>"
            f'<hr><ul><li><a href="..">..</a></li>'
        )
        for entry in sorted(
            fs.listdir(path), key=lambda x: (x["type"], x["name"], x["size"])
        ):
            # Directory then Files, then sorted lexically, then size
            name = entry["name"]

            nodename = name.split("/")[-1]
            linkname = name
            displayname = nodename
            extra_tags = " ".join([f'{k}="{v}"' for k, v in entry.items()])
            if fs.isdir(name):
                linkname += "/"
            else:
                displayname += f" ({filesize(entry['size'])})"
            r += f'<li><a {extra_tags} href="{linkname}">{displayname}</a></li>'
        r += "</ul><hr>"
        r += f"<i>sssekai v{__version__} running on Python {sys.version}</i><br>"
        r += f"<i>{fs.cache}</i><br>"
        r += "<i>page rendered in %.3fms, server time: %s</i>" % (
            (time.time() - t0) * 1000,
            datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
        )
        r += "</body></html>"
        encoded = r.encode(self.ENCODING, "surrogateescape")
        return encoded

    def handle_path(self, path):
        pass

    def do_GET(self):
        path = self.path.rstrip("/")
        if not fs.exists(path):
            self.send_error(404, "File not found")
            return
        else:
            if fs.isfile(path):
                self.send_response(200)
                self.send_header("Content-type", "application/octet-stream")
                # XXX: Size reported by bundles' metadata is not accurate.
                # If a wrong size is reported, the browser will reject the download.
                # self.send_header("Content-Length", fs.stat(path)["size"])
                self.end_headers()
                with fs.open(path, "rb") as f:
                    copyfileobj(f, self.wfile)
            else:
                listing = self.format_listing(path)
                self.send_response(200)
                self.send_header("Content-type", "text/html")
                self.send_header("Content-Length", len(listing))
                self.end_headers()
                self.wfile.write(listing)


def main_abserve(args):
    global fs
    import fsspec

    db_path = os.path.expanduser(os.path.normpath(args.db))
    fs = fsspec.filesystem("abcache", fo=db_path)
    if args.proxy:
        logger.info("Overriding proxy: %s", args.proxy)
        fs.cache.proxies = {"http": args.proxy, "https": args.proxy}
    if args.fuse:
        import fsspec.fuse

        fsspec.fuse.run(fs, "", args.fuse)
    else:
        with ThreadingHTTPServer(
            (args.host, args.port), AbServeHTTPRequestHandler
        ) as httpd:
            try:
                host, port = httpd.socket.getsockname()[:2]
                url_host = f"[{host}]" if ":" in host else host
                logger.info(
                    f"Serving HTTP on {host} port {port} "
                    f"> http://127.0.0.1:{port}/"
                    f"> http://{url_host}:{port}/"
                    f" (visit any of these URLs in your browser)"
                    f"Press Ctrl-C to stop."
                )
                httpd.serve_forever()
            except Exception as e:
                logger.info("Exiting. %s" % e)
                return 1



from io import BytesIO
from typing import BinaryIO, Iterator

SEKAI_AB_MAGIC = b"\x10\x00\x00\x00"


def decrypt_header_inplace(header: bytearray):
    assert len(header) == 128
    for i in range(0, 128, 8):
        for j in range(5):
            header[i + j] = ~header[i + j] & 0xFF
    return header


def decrypt_iter(next_bytes: callable, block_size=65536):
    header = next_bytes(4)
    if header == SEKAI_AB_MAGIC:
        header = bytearray(next_bytes(128))
        header = decrypt_header_inplace(header)
    assert len(header) <= block_size, "impossible to satisfiy bs=%d" % block_size
    header_res = block_size - len(header)
    if header_res:
        header += next_bytes(header_res)
    yield header
    while block := next_bytes(block_size):
        yield block



import os, re, json, time
from sssekai.abcache import AbCache, AbCacheEntry, logger, REGION_JP_EN, REGION_ROW
from sssekai.abcache.fs import AbCacheFilesystem, AbCacheFile
from concurrent.futures import ThreadPoolExecutor
from requests import Session
from tqdm import tqdm

class AbCacheDownloader(ThreadPoolExecutor):
    session: Session
    progress: tqdm = None

    def _ensure_progress(self):
        if not self.progress:
            self.progress = tqdm(
                total=0,
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
            )

    def _download(self, args):
        if self._shutdown:
            return
        self._ensure_progress()
        RETRIES = 1
        src, dest = args
        src: AbCacheFile
        dest: str
        tmp_dest = dest + ".tmp"
        for _ in range(0, RETRIES):
            n_written = 0
            try:
                if os.path.dirname(dest):
                    os.makedirs(os.path.dirname(dest), exist_ok=True)                
                with open(tmp_dest, "wb") as f:
                    while block := src.read(65536):
                        n_block = f.write(block)
                        self.progress.update(n_block)
                        n_written += n_block
                        if self._shutdown:
                            self.progress.update(-n_written)
                            return
                os.replace(tmp_dest, dest)
                return
            except Exception as e:
                logger.error("While downloading %s : %s. Retrying" % (src.path, e))
                self.progress.update(-n_written)
                os.remove(tmp_dest) if os.path.exists(tmp_dest) else None
                time.sleep(1)                

        if _ == RETRIES - 1:
            logger.critical("Did not download %s" % src.path)
        self._ensure_progress()

    queue: list

    def __init__(self, session, **kw) -> None:
        self.session = session
        self.queue = []
        super().__init__(**kw)

    def __enter__(self):
        return super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return super().__exit__(exc_type, exc_val, exc_tb)

    def add_link(self, file: AbCacheFile, dest: str):
        self._ensure_progress()
        self.progress.total += file.size
        return self.queue.append((file, dest))

    def run_until_complete(self):
        for _ in self.map(self._download, self.queue):
            pass


def dump_dict_by_keys(d: dict, dir: str, keep_compact: bool):
    for k, v in tqdm(d.items(), unit="file"):
        if not keep_compact and k.startswith("compact"):
            # NOTE: These are emprical and have yet been backed by decompilation
            # Assumptions:
            #   - List[Mapping[K,V]] are compacted into Mapping[K,List[V]]
            #   - Mapping[K,V] does not contain recusive mappings
            #   - Enums can only be Values and they would be stored as strings. Once compacted
            #     To 0-indexed array keyed by their parent's key in a special __ENUM__ LUT
            logger.debug("Decompacting %s", k)
            enums = v.get("__ENUM__", {})
            struct = [kk for kk in v if isinstance(v[kk], list)]
            count = len(v[struct[-1]])
            getkey = lambda kk, i: enums[kk][v[kk][i] or 0] if kk in enums else v[kk][i]
            v = [{kk: getkey(kk, i) for kk in struct} for i in range(count)]
            k = k[len("compact") :]
            k = k[0].lower() + k[1:]  # Save in the style of JP server's keys
        save_as = k + ".json"
        with open(os.path.join(dir, save_as), "w", encoding="utf-8") as f:
            json.dump(v, f, indent=4, ensure_ascii=False)
            logger.info("Saved %s", save_as)


def main_abcache(args):
    db_path = os.path.expanduser(args.db)
    cache: AbCache = AbCache()
    try:
        with open(db_path, "rb") as f:
            cache = AbCache.from_file(f)
    except Exception as e:
        logger.error("Failed to load cache from %s: %s", db_path, e)
        logger.warning("Force rebuilding cache from scratch.")
        if args.no_update:
            logger.error(
                "no_update is specified, but no valid cache file is found. AbCache will not exit."
            )
            return

    config = cache.config

    if not args.no_update:
        if config.app_region != args.app_region:
            logger.warning("Region differs. (%s->%s) Resetting database." % (config.app_region, args.app_region))
            config.app_region = args.app_region
            cache.database.reset()
        config.app_version = args.app_version
        config.app_platform = args.app_platform        
        config.app_hash = args.app_appHash
        config.row_ab_version = args.app_abVersion
        config.asset_hash = args.app_asset_hash
        config.asset_host = args.app_asset_host
        config.asset_version = args.app_asset_version
        config.auth_credential = args.auth_credential

    if args.proxy:
        logger.info("Overriding proxy: %s", args.proxy)
        cache.proxies = {"http": args.proxy, "https": args.proxy}

    if args.download_filter_cache_diff:
        diff_path = os.path.abspath(os.path.expanduser(args.download_filter_cache_diff))
        curr_path = os.path.abspath(os.path.expanduser(db_path))
        assert (
            diff_path != curr_path
        ), "Cache diff path must be different from current cache path! Make a copy of the diff cache if this is required."

    def try_auth():
        try:
            if not config.auth_available:
                logger.warning("No *valid* auth info provided.")
                # Register as anonymous user in this case
                if config.app_region in REGION_JP_EN:
                    from sssekai.abcache.auth import register_as_anonymous_user_sega

                    logger.warning("Registering as an anonymous user on SEGA servers.")
                    register_as_anonymous_user_sega(cache)
                else:
                    logger.warning("No *valid* auth info provided for ROW region.")
                    logger.warning(
                        "Anonymous user registration is not supported for those regions. You may encounter errors."
                    )
            else:
                if config.auth_available:
                    logger.info(
                        "Using cached auth credential. UserId=%s" % cache.SEKAI_USERID
                    )
            return config.auth_available
        except Exception as e:
            logger.error("Failed to authenticate: %s", e)
            return False


    if config.app_region in {"jp"}:
        cache.update_signatures()

    if args.dump_master_data:
        if config.app_region in REGION_JP_EN:
            assert (
                try_auth()
            ), "Cannot dump master data without valid auth info in EN/JP servers."
        cache.update_client_headers()
        master_data_path = os.path.expanduser(args.dump_master_data)
        os.makedirs(master_data_path, exist_ok=True)
        logger.info("Dumping master data to %s", master_data_path)
        for url in tqdm(cache.SEKAI_API_MASTER_SUITE_URLS, unit="file"):
            resp = cache.request_packed("GET", url)
            dump_dict_by_keys(
                cache.response_to_dict(resp), master_data_path, args.keep_compact
            )
        return

    if args.dump_user_data:
        assert try_auth(), "Cannot dump user data without valid auth info."
        cache.update_client_headers()
        user_data_path = os.path.expanduser(args.dump_user_data)
        os.makedirs(user_data_path, exist_ok=True)
        logger.info("Dumping user data to %s", user_data_path)
        resp = cache.request_packed("GET", cache.SEKAI_API_USER_SUITE)
        dump_dict_by_keys(
            cache.response_to_dict(resp), user_data_path, args.keep_compact
        )
        return

    try:
        if not args.no_update:
            if config.need_client_header_update:
                if config.app_region in REGION_JP_EN:
                    assert (
                        try_auth()
                    ), "Cannot update client headers without auth info. NOTE: You can fill in the EN/JP override fields (`--app-asset-...`) to bypass auth."
                cache.update_client_headers()
            cache.update_abcache_index()
            if os.path.dirname(db_path):
                os.makedirs(os.path.dirname(db_path), exist_ok=True)
            with open(db_path, "wb") as f:
                cache.save(f)
    except Exception as e:
        logger.warning("Cache update failure: %s", e)        
        logger.warning("Continuing with possibly stale cache. To explicitly do this, use --no-update.")

    if args.dump:
        dump_path = os.path.expanduser(args.dump)
        dump_path_dir = os.path.dirname(dump_path)
        if dump_path_dir and not os.path.exists(dump_path_dir):
            os.makedirs(dump_path_dir, exist_ok=True)
        logger.info("Dumping AssetBundle index to %s", dump_path)
        from dataclasses import asdict
        with open(dump_path, "w", encoding="utf-8") as f:
            json.dump(asdict(cache.abcache_index), f, indent=4, ensure_ascii=False)
        
    if args.download_dir:
        download_dir = os.path.expanduser(args.download_dir)
        bundles = set()

        def _iter_bundles():
            filter_pred = []
            if args.download_filter:
                logger.info(
                    "Filtering bundles with regex pattern: %s", args.download_filter
                )
                pattern = re.compile(args.download_filter)
                filter_pred.append(lambda bundle: pattern.match(bundle.bundleName))
            if args.download_filter_cache_diff:
                logger.info("Filtering bundles with cache diff")
                diff_path = args.download_filter_cache_diff
                diff_path = os.path.abspath(os.path.expanduser(diff_path))
                logger.info("Loading cache diff from %s", diff_path)
                diff_cache = AbCache.from_file(open(diff_path, "rb"))

                def __diff_pred(bundle: AbCacheEntry):
                    diff = diff_cache.abcache_index.bundles.get(bundle.bundleName, None)
                    if diff is not None:
                        if diff.hash != bundle.hash:
                            return True
                        else:
                            logger.debug(
                                "Skipped %s due to cache diff hit (hash)",
                                bundle.bundleName,
                            )
                            return False
                    else:
                        return True

                filter_pred.append(__diff_pred)
            if filter_pred:
                for name, entry in cache.abcache_index.bundles.items():
                    if all(pred(entry) for pred in filter_pred):
                        yield name
            else:
                logger.warning(
                    "No filter pattern specified. All bundles will be downloaded."
                )
                for name, entry in cache.abcache_index.bundles.items():
                    yield name

        for bundleName in _iter_bundles():
            bundles.add(bundleName)
        basebundles = bundles.copy()
        logger.info("Selected %d bundles to download", len(basebundles))
        if args.download_ensure_deps:
            for bundleName in basebundles:
                cache.get_or_update_dependency_tree_flatten(bundleName, bundles)
            logger.info("Added dependencies:")
            for dep in bundles - basebundles:
                logger.info("   - %s", dep)
        fs = AbCacheFilesystem(cache_obj=cache)
        with AbCacheDownloader(fs, max_workers=args.download_workers) as downloader:
            logger.info("Downloading %d bundles to %s" % (len(bundles), download_dir))
            for bundleName in bundles:
                dst = os.path.join(download_dir, bundleName)                
                if args.download_no_overwrite and os.path.exists(dst):
                    logger.info("Skipping existing bundle %s", bundleName)
                else:
                    downloader.add_link(
                        fs.open(bundleName), os.path.join(download_dir, bundleName)
                    )
            try:
                downloader.run_until_complete()
            except KeyboardInterrupt:
                logger.warning("Download interrupted by user.")
                downloader.shutdown(wait=False, cancel_futures=True)
                return 1
    return 0



from sssekai.unity.AnimationClip import AnimationHelper, Interpolation
from UnityPy.classes import AnimationClip
from logging import getLogger

logger = getLogger(__name__)


# Thanks! https://github.com/Perfare/UnityLive2DExtractor/blob/master/UnityLive2DExtractor/CubismMotion3Converter.cs
def to_motion3(
    helper: AnimationHelper, crc_table: dict, raw_clip: AnimationClip = None
) -> dict:
    """Convert AnimationHelper instance to Live2D Motion3 format

    Args:
        helper (AnimationHelper): AnimationHelper instance
        crc_table (dict): CRC table for path binding
        raw_clip (AnimationClip, optional): Raw AnimationClip instance. Used with custom Events. Defaults to None.

    Returns:
        dict: Live2D Motion3 data
    """
    motion = {
        "Version": 3,
        "Meta": {
            "Name": helper.Name,
            "Duration": helper.Duration,
            "Fps": helper.SampleRate,
            "Loop": True,
            "AreBeziersRestricted": True,
            "CurveCount": 0,
            "UserDataCount": 0,
            "TotalPointCount": 0,
            "TotalSegmentCount": 0,
            "TotalUserDataSize": 0,
        },
        "Curves": [],
        "UserData": [],
    }
    floatCurves = list(helper.FloatCurves)
    motion["Meta"]["CurveCount"] = len(floatCurves)
    for curve in floatCurves:
        segments = list()
        segments.append(0)
        segments.append(curve.Data[0].value)
        for key in curve.Data[1:]:
            ipo = key.interpolation_segment(key.prev, key)[0]
            match ipo:
                case Interpolation.Constant:
                    segments.append(2)  # SteppedSegment
                    segments.append(key.time)
                    segments.append(key.value)
                    motion["Meta"]["TotalPointCount"] += 1
                case Interpolation.Stepped:
                    segments.append(2)  # SteppedSegment
                    segments.append(key.time)
                    segments.append(key.value)
                    motion["Meta"]["TotalPointCount"] += 1
                case Interpolation.Hermite:
                    lhs, rhs = key.prev, key
                    dx = (rhs.time - lhs.time) / 3
                    # 1/3rd rule applies to the editor as well
                    segments.append(1)  # BezierSegment
                    segments.append(lhs.time + dx)
                    segments.append(lhs.outSlope * dx + lhs.value)
                    segments.append(rhs.time - dx)
                    segments.append(rhs.value - rhs.inSlope * dx)
                    segments.append(rhs.time)
                    segments.append(rhs.value)
                    motion["Meta"]["TotalPointCount"] += 3
                case Interpolation.Linear:
                    segments.append(0)  # LinearSegment
                    segments.append(key.time)
                    segments.append(key.value)
                    motion["Meta"]["TotalPointCount"] += 1
            motion["Meta"]["TotalSegmentCount"] += 1
        path = curve.Path
        if path in crc_table:
            target, id = crc_table[path].split("/")
            if target == "Parameters":
                target = "Parameter"
            if target == "Parts":
                target = "PartOpacity"
        else:
            logger.warning("Failed to bind path CRC %s to any Live2D path" % path)
            target, id = "PartOpacity", str(path)
        motion["Curves"].append({"Target": target, "Id": id, "Segments": segments})
    if raw_clip:
        for event in raw_clip.m_Events:
            motion["UserData"].append({"time": event.time, "value": event.data})
            motion["Meta"]["UserDataCount"] += 1
            motion["Meta"]["TotalUserDataSize"] += len(event.data)
    return motion



import struct
from collections import defaultdict
from typing import Dict, List, Generator
from bisect import bisect_right
from dataclasses import dataclass, field
from UnityPy.enums import ClassIDType
from UnityPy.classes import (
    AnimationClip,
    StreamedClip,
    GenericBinding,
    # ---
    Vector3Curve,
    QuaternionCurve,
    FloatCurve,
)
from UnityPy.classes.math import Vector3f as Vector3, Quaternionf as Quaternion
from UnityPy.streams.EndianBinaryReader import EndianBinaryReader
from logging import getLogger
from enum import IntEnum
from zlib import crc32

logger = getLogger(__name__)


def num_floats(value):
    if type(value) == Vector3:
        return 3
    elif type(value) == Quaternion:
        return 4
    else:
        return 1


def as_floats(value):
    if type(value) == Vector3:
        return [value.x, value.y, value.z]
    elif type(value) == Quaternion:
        return [value.x, value.y, value.z, value.w]
    else:
        return [value]


def from_floats(*values):
    if len(values) == 1:
        return values[0]
    elif len(values) == 3:
        return Vector3(*values)
    else:
        return Quaternion(*values)


# GenericBinding Attributes
kBindTransformPosition = 1
kBindTransformRotation = 2
kBindTransformScale = 3
kBindTransformEuler = 4
# ---
kAttributeSizeof = lambda x: {
    kBindTransformPosition: 3,
    kBindTransformRotation: 4,  # Quaternion
    kBindTransformScale: 3,
    kBindTransformEuler: 3,
}.get(x, 1)


# Interpolation Types
class Interpolation(IntEnum):
    Hermite = 1
    Linear = 2
    Stepped = 3
    Constant = 4


class StreamedClipKey:
    """
    dx = rhs.time - lhs.time;
    dx = max(dx, 0.0001F);
    dy = rhs.value - lhs.value;
    length = 1.0F / (dx * dx);

    m1 = lhs.outSlope;
    m2 = rhs.inSlope;
    d1 = m1 * dx;
    d2 = m2 * dx;

    cache.coeff[0] = (d1 + d2 - dy - dy) * length / dx;
    cache.coeff[1] = (dy + dy + dy - d1 - d1 - d2) * length;
    cache.coeff[2] = m1;
    cache.coeff[3] = lhs.value;
    """

    index: int

    coeff: List[float]
    int_coeff: List[int]

    raw_coeff: bytes

    time: float
    prev: "StreamedClipKey" = None

    def __init__(self, reader: EndianBinaryReader, time: float):
        self.index = reader.read_int()
        endian = reader.endian

        self.raw_coeff = reader.read_bytes(4 * 4)
        self.coeff = struct.unpack(endian + "4f", self.raw_coeff)
        self.int_coeff = struct.unpack(endian + "4i", self.raw_coeff)

        self.time = time

    inSlope: float = float("inf")

    @property
    def outSlope(self):
        return self.coeff[2]

    @property
    def value(self):
        return self.coeff[3]

    @property
    def int_value(self):
        return self.int_coeff[3]

    def calc_next_in_slope(self, dx: float, rhs):
        # Stepped
        if self.coeff[0] == 0 and self.coeff[1] == 0 and self.coeff[2] == 0:
            return float("inf")

        dx = max(dx, 0.0001)
        dy = rhs.value - self.value
        length = 1.0 / (dx * dx)
        d1 = self.outSlope * dx
        d2 = dy + dy + dy - d1 - d1 - self.coeff[1] / length
        return d2 / dx


class StreamedClipFrame:
    time: float
    keys: List[StreamedClipKey]

    def __init__(self, reader: EndianBinaryReader):
        self.time = reader.read_float()
        self.keys = [
            StreamedClipKey(reader, self.time) for _ in range(reader.read_int())
        ]


def read_streamed_clip_frames(clip: StreamedClip) -> List[StreamedClipFrame]:  # O(n)
    frames: List[StreamedClipFrame] = []
    buffer = b"".join(val.to_bytes(4, "big") for val in clip.data)
    reader = EndianBinaryReader(buffer)
    while reader.Position < reader.Length:
        frames.append(StreamedClipFrame(reader))
    preKeys: Dict[int, StreamedClipKey] = dict()
    for frame in frames:
        for curveKey in frame.keys:
            preKey = preKeys.get(curveKey.index, None)
            if preKey:
                curveKey.inSlope = preKey.calc_next_in_slope(
                    frame.time - preKey.time, curveKey
                )
            preKeys[curveKey.index] = curveKey
    return frames


@dataclass
class KeyframeHelper:
    """Custom KeyFrame container/lookup type analogus to Unity's Keyframe
    whilst providing dynamic type support (float/int/Euler/Vector/Quaternion) and interpolation
    """

    time: float
    typeID: int
    value: float | Vector3 | Quaternion = 0

    isDense: bool = False
    isConstant: bool = False
    # Used for Hermite curves
    inSlope: float | Vector3 | Quaternion = 0
    outSlope: float | Vector3 | Quaternion = 0

    prev: "KeyframeHelper" = None
    next: "KeyframeHelper" = None

    endian: str = ">"

    @property
    def int_value(self):
        assert type(self.value) == float
        b = struct.pack("<f", self.value)
        return struct.unpack(self.endian + "i", b)[0]

    def __repr__(self):
        return f"KeyFrame({self.time}, {self.value}, inSlope={self.inSlope}, outSlope={self.outSlope})"

    @staticmethod
    def interpolate_cubic_hermite_unit(t, p0, m0, m1, p1):
        # https://en.wikipedia.org/wiki/Cubic_Hermite_spline
        t2 = t * t
        t3 = t2 * t
        a = 2 * t3 - 3 * t2 + 1
        b = t3 - 2 * t2 + t
        c = -2 * t3 + 3 * t2
        d = t3 - t2
        return a * p0 + b * m0 + c * p1 + d * m1

    @staticmethod
    def interpolate_linear_unit(t, p0, p1):
        return p0 + t * (p1 - p0)

    @staticmethod
    def interpolate_stepped(t, p0, p1):
        return p0

    @staticmethod
    def interpolation_segment(
        lhs: "KeyframeHelper", rhs: "KeyframeHelper"
    ) -> List[Interpolation]:
        """Interpolation of the segment between lhs and rhs"""
        if lhs.isDense:
            ipo = [Interpolation.Linear] * num_floats(lhs.value)
        elif lhs.isConstant or not rhs:
            ipo = [Interpolation.Constant] * num_floats(lhs.value)
        else:
            ipo = [Interpolation.Hermite] * num_floats(lhs.value)
            for i, (lhsInSlope, lhsOutSlope, rhsInSlope, rhsOutSlope) in enumerate(
                zip(
                    as_floats(lhs.inSlope),
                    as_floats(lhs.outSlope),
                    as_floats(rhs.inSlope),
                    as_floats(rhs.outSlope),
                )
            ):
                if any((x == float("inf") for x in (lhsOutSlope, rhsInSlope))):
                    ipo[i] = Interpolation.Stepped
                else:
                    ipo[i] = Interpolation.Hermite
        return ipo

    @staticmethod
    def interpolate(
        t, lhs: "KeyframeHelper", rhs: "KeyframeHelper", unit_t: bool = False
    ) -> float | Vector3 | Quaternion:
        EPS = 1e-8
        if not rhs:
            return lhs.value
        dx = rhs.time - lhs.time        
        if dx < EPS:
            return lhs.value
        if not unit_t:
            t -= lhs.time
            t /= dx  # Normalized to [0,1]
        result = [0] * num_floats(lhs.value)
        for i, (
            lhsValue,
            rhsValue,
            lhsInSlope,
            lhsOutSlope,
            rhsInSlope,
            rhsOutSlope,
            interpolation,
        ) in enumerate(
            zip(
                as_floats(lhs.value),
                as_floats(rhs.value),
                as_floats(lhs.inSlope),
                as_floats(lhs.outSlope),
                as_floats(rhs.inSlope),
                as_floats(rhs.outSlope),
                KeyframeHelper.interpolation_segment(lhs, rhs),
            )
        ):
            match interpolation:
                case Interpolation.Hermite:
                    result[i] = KeyframeHelper.interpolate_cubic_hermite_unit(
                        t, lhsValue, lhsOutSlope * dx, rhsInSlope * dx, rhsValue
                    )
                case Interpolation.Linear:
                    # Lerp doesn't seem to be explicitly used as a cubic hermite curve when correctly
                    # setup can produce results of a linear interpolation
                    # However this is necessary for Dense curves. So we'll keep it here.
                    result[i] = KeyframeHelper.interpolate_linear_unit(
                        t, lhsValue, rhsValue
                    )
                    # result[i] = KeyFrame.interpolate_cubic_hermite_unit(
                    #     t, lhsValue, lhsOutSlope * dx, rhsInSlope * dx, rhsValue
                    # )
                case Interpolation.Stepped:
                    result[i] = KeyframeHelper.interpolate_stepped(
                        t, lhsValue, rhsValue
                    )
                case Interpolation.Constant:
                    result[i] = lhsValue
        return result[0] if len(result) == 1 else from_floats(*result)


def _read_clip_keyframe(
    time: float,
    binding: GenericBinding,
    keys: Generator[StreamedClipKey | KeyframeHelper, None, None],
) -> KeyframeHelper:
    result = KeyframeHelper(time, binding.typeID)
    if binding.typeID == ClassIDType.Transform:
        dimension = kAttributeSizeof(binding.attribute)
        values, outSlopes, inSlopes = [0] * dimension, [0] * dimension, [0] * dimension
        for i in range(dimension):
            key = next(keys)
            values[i] = key.value
            outSlopes[i] = key.outSlope
            inSlopes[i] = key.inSlope
        result.value = from_floats(*values)
        result.outSlope = from_floats(*outSlopes)
        result.inSlope = from_floats(*inSlopes)
    else:
        if binding.isIntCurve:
            key = next(keys)
            result.value = key.int_value
            # Can only be stepped
            result.outSlope = float("inf")
            result.inSlope = float("inf")
        elif binding.isPPtrCurve:
            # !! TODO PPtr curves not yet implemented
            pass
        else:
            # Default to float curves
            key = next(keys)
            result.value = key.value
            result.outSlope = key.outSlope
            result.inSlope = key.inSlope
    return result


@dataclass
class CurveHelper:
    """Custom Curve container/lookup type analogus to Unity's AnimationCurve

    Note:
        - Properties here aren't prefixed with `m_` like in Unity's AnimationClip
        - For more information, read docs on `sssekai.unity.KeyFrame`
    """

    Binding: GenericBinding
    Data: List[KeyframeHelper] = field(default_factory=list)

    def evaluate(self, t: float) -> float | Vector3 | Quaternion:  # O(log n)
        lhs = bisect_right(self.Data, t, key=lambda x: x.time) - 1
        lhs = max(lhs, 0)
        lhs = self.Data[lhs]
        rhs = lhs.next
        return KeyframeHelper.interpolate(t, lhs, rhs)

    def resample_dense(self, times: List[float]) -> "CurveHelper": # O(n)
        """Resamples the curve at given times **as a Dense curve**, losing all tangent information.

        This guarantees accuracy when played back at a constant sample rate at the cost of space
        and interpolation information.

        Args:
            times (List[float]): Times to sample the curve at

        Returns:
            CurveHelper: A new CurveHelper with the sampled values
        """
        j = 0
        data = [None] * len(times)
        for i, t in enumerate(times):
            while j + 1 < len(self.Data) and self.Data[j].time <= t:
                j += 1            
            lhs = max(j - 1, 0)            
            data[i] = KeyframeHelper.interpolate(t, self.Data[lhs], self.Data[lhs].next)
        slope = from_floats(*([0] * num_floats(data[0])))
        data = [KeyframeHelper(t, self.Binding.typeID, value, isDense=True, inSlope=slope, outSlope=slope) for t, value in zip(times, data)]
        for i in range(1, len(data)):
            data[i].prev = data[i - 1]
            data[i - 1].next = data[i]
        return CurveHelper(self.Binding, data)

    @property
    def Duration(self):
        return self.Data[-1].time - self.Data[0].time

    @property
    def Path(self):
        return self.Binding.path

    @property
    def Attribute(self):
        return self.Binding.attribute

    @property
    def is_transform_curve(self):
        return self.Binding.typeID == ClassIDType.Transform


@dataclass
class AnimationHelper:
    """Custom Animation container/lookup type analogus to Unity's AnimationClip

    Note:
        - Properties here aren't prefixed with `m_` like in Unity's AnimationClip
        - Mapping of functionality isn't 1:1 with Unity's AnimationClip since
          there's no access to post-build clip data. Everything here is ready for use.
        - For more information, read docs on `sssekai.unity.Curve` and `sssekai.unity.KeyFrame`
    """

    Name: str
    Duration: float
    SampleRate: float
    # Dict[internal hash, Curve]
    RawCurves: Dict[int, CurveHelper] = field(default_factory=dict)
    # Curves[Attribute][Path Hash] = Curve
    Curves: Dict[int, Dict[int, CurveHelper]] = field(
        default_factory=lambda: defaultdict(dict)
    )
    # CurvesT[Path Hash][Attribute] = Curve
    CurvesT: Dict[int, Dict[int, CurveHelper]] = field(
        default_factory=lambda: defaultdict(dict)
    )
    # From path id to the full path
    InvCRC: Dict[int, str] = field(default_factory=dict)

    @staticmethod
    def hash_of(binding: GenericBinding):
        return hash((binding.path, binding.attribute))

    def get_curve(self, binding: GenericBinding):
        hs = self.hash_of(binding)
        if not hs in self.RawCurves:
            curve = CurveHelper(binding)
            # Refcounted
            self.RawCurves[hs] = curve
            self.Curves[binding.attribute][binding.path] = curve
            self.CurvesT[binding.path][binding.attribute] = curve
            return curve
        else:
            return self.RawCurves[hs]

    # Shorthands for Unity-like access
    @property
    def TransformCurves(self):
        return filter(lambda c: c.is_transform_curve, self.RawCurves.values())

    @property
    def FloatCurves(self):
        return filter(lambda c: not c.is_transform_curve, self.RawCurves.values())

    @property
    def PositionCurves(self):
        return self.Curves[kBindTransformPosition].values()

    @property
    def RotationCurves(self):
        return self.Curves[kBindTransformRotation].values()

    @property
    def ScaleCurves(self):
        return self.Curves[kBindTransformScale].values()

    @property
    def EulerCurves(self):
        return self.Curves[kBindTransformEuler].values()

    @property
    def PPtrCurves(self):
        raise NotImplementedError

    def process_legacy(self, src: AnimationClip) -> "AnimationHelper":
        """Reads pre-build AnimationClip data and converts it to an Animation

        NOTE: This only works with Legacy AnimationClips, where the animation data
        is not compacted into a MuscleClip and is instead stored in the AnimationClip itself.

        Note:
            `Animation` in this context is NOT a part of Unity (which coincidentally also has an `Animation` class)
            Read `sssekai.unity.Animation`'s docs for more information

        Args:
            src (AnimationClip): An Unity AnimationClip object

        Returns:
            sssekai.unity.Animation
        """

        def process_curves(
            curves: List[Vector3Curve | QuaternionCurve | FloatCurve], attribute: int
        ):
            for curve in curves:
                binding = GenericBinding(
                    attribute or curve.attribute,
                    None,
                    False,
                    crc32(curve.path.encode("utf-8")),
                    None,
                )
                binding.typeID = ClassIDType.Transform
                self.InvCRC[binding.path] = curve.path
                keyframes = [
                    KeyframeHelper(
                        time=key.time,
                        typeID=binding.typeID,
                        value=key.value,
                        inSlope=key.inSlope,
                        outSlope=key.outSlope,
                    )
                    for key in curve.curve.m_Curve
                ]
                self.get_curve(binding).Data = keyframes
            pass

        process_curves(src.m_PositionCurves, kBindTransformPosition)
        process_curves(src.m_RotationCurves, kBindTransformRotation)
        process_curves(src.m_ScaleCurves, kBindTransformScale)
        process_curves(src.m_EulerCurves, kBindTransformEuler)
        process_curves(src.m_FloatCurves, None)
        # ! TODO: PPtr curves not yet supported
        # process_curves(src.m_PPtrCurves, ClassIDType.PPtr)
        for curve in self.RawCurves.values():
            for i in range(1, len(curve.Data)):
                curve.Data[i].prev = curve.Data[i - 1]
                curve.Data[i - 1].next = curve.Data[i]
        return self

    def process(self, src: AnimationClip) -> "AnimationHelper":
        """Reads post-build AnimationClip data and converts it to an Animation

        Note:
            `Animation` in this context is NOT a part of Unity (which coincidentally also has an `Animation` class)
            Read `sssekai.unity.Animation`'s docs for more information

        Args:
            src (AnimationClip): An Unity AnimationClip object

        Returns:
            sssekai.unity.Animation
        """
        mClip = src.m_MuscleClip.m_Clip.data
        mClipBinding = src.m_ClipBindingConstant
        mClipBindingCurveSizesPfx = [
            kAttributeSizeof(b.attribute) for b in mClipBinding.genericBindings
        ]
        for i in range(1, len(mClipBindingCurveSizesPfx)):
            mClipBindingCurveSizesPfx[i] += mClipBindingCurveSizesPfx[i - 1]

        def mClipFindBinding(cur) -> GenericBinding:  # O(log n)
            index = bisect_right(mClipBindingCurveSizesPfx, cur)
            index = max(index, 0)
            index = min(index, len(mClipBinding.genericBindings) - 1)
            return mClipBinding.genericBindings[index]

        def mClipGetNextCurve(
            cur, keys
        ) -> int:  # O(1) since curve counts are between 1 and 4
            i = cur
            while i < len(keys) and keys[cur].index == keys[i].index:
                i += 1
            return i

        # https://docs.unity3d.com/Manual/class-Animator.html
        # When actually stored post build, the data is *only* stored with Stream, Dense or Constant clips
        # Float, TRS curves will be optimized into these formats and will not be available.

        # StreamClip
        # Animation is stored with Keyframe Reduction
        # Interpolated with Hermite curves
        mStreamClipFrames = read_streamed_clip_frames(mClip.m_StreamedClip)
        for frame in mStreamClipFrames:
            curveIndex = 0

            def __keygen():
                nonlocal curveIndex
                while True:
                    currIndex = curveIndex
                    curveIndex = mClipGetNextCurve(curveIndex, frame.keys)
                    yield frame.keys[currIndex]

            __keygen_g = __keygen()
            while curveIndex < len(frame.keys):
                binding = mClipFindBinding(frame.keys[curveIndex].index)
                curve = self.get_curve(binding)
                key = _read_clip_keyframe(frame.time, binding, __keygen_g)
                curve.Data.append(key)

        # DenseClip
        # Animation is stored as a dense clip
        # Values are to be interpolated linearly
        m_DenseClip = mClip.m_DenseClip
        curveOffset = mClip.m_StreamedClip.curveCount
        for frameIndex in range(0, m_DenseClip.m_FrameCount):
            time = m_DenseClip.m_BeginTime + frameIndex / m_DenseClip.m_SampleRate
            frameOffset = frameIndex * m_DenseClip.m_CurveCount
            curveIndex = 0

            def __keygen():
                nonlocal curveIndex
                while True:
                    currIndex = curveIndex
                    curveIndex += 1
                    yield KeyframeHelper(
                        time,
                        binding.typeID,
                        m_DenseClip.m_SampleArray[frameOffset + currIndex],
                    )

            __keygen_g = __keygen()
            while curveIndex < m_DenseClip.m_CurveCount:
                binding = mClipFindBinding(curveOffset + curveIndex)
                curve = self.get_curve(binding)
                key = _read_clip_keyframe(time, binding, __keygen_g)
                key.isDense = True
                curve.Data.append(key)

        # ConstantClip
        # Unchanging values.
        m_ConstantClip = mClip.m_ConstantClip
        curveOffset = mClip.m_StreamedClip.curveCount + mClip.m_DenseClip.m_CurveCount
        time = 0
        for _ in range(0, 2):  # first and last frame
            curveIndex = 0

            def __keygen():
                nonlocal curveIndex
                while True:
                    currIndex = curveIndex
                    curveIndex += 1
                    yield KeyframeHelper(
                        time,
                        binding.typeID,
                        m_ConstantClip.data[currIndex],
                    )

            __keygen_g = __keygen()
            while curveIndex < len(m_ConstantClip.data):
                binding = mClipFindBinding(curveOffset + curveIndex)
                curve = self.get_curve(binding)
                key = _read_clip_keyframe(time, binding, __keygen_g)
                key.isConstant = True
                curve.Data.append(key)
                binding = mClipFindBinding(curveOffset + curveIndex)
            time = src.m_MuscleClip.m_StopTime

        for curve in self.RawCurves.values():
            for i in range(1, len(curve.Data)):
                curve.Data[i].prev = curve.Data[i - 1]
                curve.Data[i - 1].next = curve.Data[i]
        return self

    @staticmethod
    def from_clip(src: AnimationClip):
        helper = AnimationHelper(
            src.m_Name, src.m_MuscleClip.m_StopTime, src.m_SampleRate
        )
        if src.m_Legacy:
            helper.process_legacy(src)
        else:
            helper.process(src)
        return helper


# Backwards compatibility
Animation = AnimationHelper
KeyFrame = KeyframeHelper
Curve = CurveHelper

vec3_quat_as_floats = as_floats
vec3_quat_from_floats = from_floats
read_animation = AnimationHelper.from_clip
